package cz.upol.foolang.tutorials;

import static cz.upol.foolang.jbc.Util.typeSignature;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;

import cz.upol.foolang.jbc.ClassBuilder;
import cz.upol.foolang.jbc.MethodBuilder;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.bytecode.Bytecode;
import javassist.bytecode.Opcode;

public class Sample04 {
	
	private static final String WORKING_DIRECTORY = System.getProperty("java.io.tmpdir");
	private static final String CLASS_NAME = "QuxClass";
	
	public static void main(String[] args) throws Exception {

		// vytvori tridu dle definice
		// class QuxClass {
		//    private int x;
		//    private int y;
		// ...
		ClassBuilder quxBuilder = new ClassBuilder("QuxClass");
	
		quxBuilder.addField(CtClass.intType, "x", Modifier.PRIVATE);
		quxBuilder.addField(CtClass.intType, "y", Modifier.PRIVATE);
		
		createConstructor(quxBuilder);
		createGetterX(quxBuilder);
		createSetterX(quxBuilder);
		
		quxBuilder.save(WORKING_DIRECTORY);
		
		testMethods();
	}
	
	/**
	 * Vytvori setter odpovidajici deklaraci
	 * public void setX(int a) {
	 *    this.x = a;
	 * }
	 */

	private static void createSetterX(ClassBuilder quxBuilder) {
		MethodBuilder setXMethod = quxBuilder.createMethod("setX", Modifier.PUBLIC, CtClass.voidType, CtClass.intType);
		Bytecode bc = setXMethod.createByteCode(2, 2);
		bc.addAload(0); // this
		bc.addIload(1); // argument
		bc.addPutfield(quxBuilder.getGeneratedClass(), "x", typeSignature(CtClass.intType));
		bc.add(Opcode.RETURN);
	}

	/**
	 * Vytvori metodu odpovidajici deklaraci:
	 * public int getX() {
	 *    return this.x;
	 * }
	 */
	private static void createGetterX(ClassBuilder quxBuilder) {
		MethodBuilder getXMethod = quxBuilder.createMethod("getX", Modifier.PUBLIC, CtClass.intType);
		Bytecode bc = getXMethod.createByteCode(1, 1);
		bc.addAload(0); // this
		bc.addGetfield(quxBuilder.getGeneratedClass(), "x", typeSignature(CtClass.intType));
		bc.add(Opcode.IRETURN);
	}
	
	/**
	 * Vytvori konstruktor odpovidaji deklaraci
	 * public QuxClass(int a, int b) {
	 *    super();
	 *    this.x = a;
	 *    this.y = b;
	 * }
	 */
	public static void createConstructor(ClassBuilder quxBuilder) throws NotFoundException {
		
		ClassPool classPool = ClassPool.getDefault();
		
		MethodBuilder constructor = quxBuilder.createConstructor(Modifier.PUBLIC, CtClass.intType, CtClass.intType);
		Bytecode bc = constructor.createByteCode(2, 3);
		bc.addAload(0); // this

		CtClass classObject = classPool.get("java.lang.Object");
		bc.addInvokespecial(classObject, "<init>", CtClass.voidType, new CtClass[] { });
		
		bc.addAload(0);
		bc.addIload(1);
		bc.addPutfield(quxBuilder.getGeneratedClass(), "x", typeSignature(CtClass.intType));
		
		bc.addAload(0);
		bc.addIload(2);
		bc.addPutfield(quxBuilder.getGeneratedClass(), "y", typeSignature(CtClass.intType));
		
		bc.add(Opcode.RETURN);
	}
	
	/**
	 * Otestuje metody vytvorene v metode {@link #main}
	 */
	public static void testMethods() throws Exception {
		URL url = new URL("file:" + WORKING_DIRECTORY + "/");
		URLClassLoader cl = new URLClassLoader(new URL[] { url });

		Class<?> c = cl.loadClass(CLASS_NAME);
		
		// vytvori objekt
		Constructor<?> constructor = c.getConstructor(int.class, int.class);
		Object testObject = constructor.newInstance(10, 20);

		// zjisti hodnotu x
		Method m1 = c.getMethod("getX");
		Object o1 = m1.invoke(testObject);
		System.out.println(o1);
		
		// zmeni hodnotu x
		Method m2 = c.getMethod("setX", int.class);
		m2.invoke(testObject, 42);

		// zjisti hodnotu x jeste jednou
		Object o2 = m1.invoke(testObject);
		System.out.println(o2);
		
		cl.close();
	}

}
